#!/usr/bin/perl
# convert L3 BASIC from internal code to ascii code
# V0.4.3
#
# Copyright (c) 2012, Sasaji
#use utf8;
use strict;
use Getopt::Std;
use File::Basename;

our %opts = ();
our $charset_type = "0";	# 
our $through_mode = 1;		# 
our $basic_type = "";

our %basic_code_table = ();
our %char_code_table = ();

our $app_path=dirname($0)."/";

getopts("di:t",\%opts);

if (exists($opts{"d"})) {
	$basic_type = "disk";
}
if ($opts{"i"} =~ /^[01]$/) {
	$through_mode = 0;
	$charset_type = $opts{"i"};
}

if (@ARGV <= 0) {
	&usage();
	exit 0;
}

if (set_basic_code_table(\%basic_code_table)) {
	exit 1;
}
if (set_char_code_table(\%char_code_table)) {
	exit 1;
}

foreach my $infile (@ARGV) {
	&main($infile);
}

# $k=getc(STDIN);

exit 0;

#####
sub usage() {
	my $prog_name=basename($0);

	print "$prog_name [options] file ...\n";
	print " Decode to ascii text from L3 BASIC internal code."."\n";
	print '  options:'."\n";
	print '  -d     : parse as "DISK BASIC".'."\n";
	print '  -i 0   : convert file(s) using interace mode charset(utf-8).'."\n";
	print '           hankaku kata/hira -> zenkaku kata/hira'."\n";
	print '  -i 1   : convert file(s) using non-interace mode charset(utf-8).'."\n";
}

#####

sub main() {
	my($infile) = @_;
	my $fh;

	my $buf="";
	my $buflen = 0;
	my $bufpos = 0;


	if (! open($fh,$infile)) {
		print STDERR $infile.' cannot open.'."\n";
		return;
	}
	binmode($fh);

	my %parse_state;
	my $val = 0;

	&init_data(\%parse_state);

	while (($buflen = read($fh,$buf,1024)) > 0) {
		for($bufpos=0; $bufpos < $buflen; $bufpos++) {
			$val = unpack("C",substr($buf, $bufpos, 1));
			$parse_state{file_pos}++;
#			printf "%02X ",$val;

			if ($parse_state{mode} >= 0) {
				&parse_data($val, \%parse_state);
			} else {
				print STDERR 'Unknown data exists.'."\n";
				last;
			}
		}
	}
	close($fh);

#	print "end\n";
}

#################
sub init_data() {
	my($rs)=@_;

	$rs->{mode} = 0;
	$rs->{file_pos} = -1;		# 

	$rs->{addr_next} = 0;		# 
	$rs->{line_num} = 0;		# 
	$rs->{subpos} = 0;			# 

	$rs->{quoted} = 0;			# 
	$rs->{data_area} = 0;		# 
	$rs->{comment_area} = 0;	# 

	$rs->{body} = "";			# 

	$rs->{commandshift} = 0;	# 

}

sub parse_data() {
	my($val, $rs)=@_;

	my $chrstr = "";
	my $cmd = "";

	# for BASIC MASTER LEVEL 3
	if ($rs->{mode} == 0) {
		# init
#				print "mode: 0\n";
		if ($val == 0xfe) {
			print STDERR 'This program is encrypted.'."\n";
		}
		if ($val != 0xfe && $val != 0xff) {
			$rs->{mode} = 1;
			$rs->{subpos} = 0;
		}
	}

	if ($rs->{mode} == 1) {
		# read next address point(2bytes)
		if ($rs->{subpos} == 0) {
			$rs->{addr_next} = $val * 256;
			$rs->{subpos}++;

		} else {
			$rs->{addr_next} += $val;

			if ($rs->{addr_next} == 0) {
				# end of data
				$rs->{mode} = -1;
				$rs->{subpos} = 0;
			} else {
				$rs->{mode} = 2;
				$rs->{subpos} = 0;
			}
		}
	} elsif ($rs->{mode} == 2) {
		# read linenumber(2bytes)
		if ($rs->{subpos} == 0) {
			$rs->{line_num} = $val * 256;
			$rs->{subpos}++;

		} elsif ($rs->{subpos} == 1) {
			$rs->{line_num} += $val;

			$rs->{mode} = 3;
			$rs->{body} = "";
			$rs->{subpos} = 0;

		}
	} elsif ($rs->{mode} == 3) {
		# body
		if ($rs->{commandshift} == 1) {
				$cmd=$basic_code_table{$val+(255*256)};
				if (ref($cmd) eq "HASH") {
					$rs->{body} .= $cmd->{val};

				} else {
#							$rs->{body} .= sprintf("[0xff%02X]",$val);
					$rs->{body} .= sprintf("%c%c",0xff,$val);

				}

			$rs->{commandshift} = 0;

		} else {
			if ($val == 0) {
				# end of line (see below)
			} elsif ($val == 0x22) {
				# "
				$rs->{body} .= sprintf("%c",$val);
				if ($rs->{data_area} != 1 && $rs->{comment_area} != 1) {
					$rs->{quoted} = 1 - $rs->{quoted};	# toggle quoted string
				}
				$cmd = "";
			} elsif ($val == 0xff) {
				# extend statement
				if ($rs->{quoted} != 1 && $rs->{data_area} != 1 && $rs->{comment_area} != 1) {
					$rs->{commandshift} = 1;
				}
				$cmd = "";
#					} elsif ($val >= 0x20 && $val < 0x7f) {
#						# 
#						$rs->{body} .= sprintf("%c",$val);
#						$cmd = "";
			} else {
				if ($rs->{quoted} == 1 || $rs->{data_area} == 1 || $rs->{comment_area} == 1) {
					# quoted string or DATA line or REM line
#								$rs->{body} .= sprintf("[0x%02X]",$val);
					if ($through_mode == 1) {
						$chrstr = "";
					} else {
						$chrstr = $char_code_table{$charset_type}{$val}{val};
					}
					if ($chrstr ne "") {
						$rs->{body} .= $chrstr;
					} else {
						$rs->{body} .= sprintf("%c",$val);
					}
				} else {
					$cmd=$basic_code_table{$val};

					if (ref($cmd) eq "HASH") {
						if ($cmd->{attr} =~ /colon/) { # ELSE
							$rs->{body} =~ s/:$//;
						}

						$rs->{body} .= $cmd->{val};

						if ($cmd->{attr} =~ /data/) { # DATA

							$rs->{data_area} = 1 - $rs->{data_area};
						} elsif ($cmd->{attr} =~ /comment/) { # ' REM

							$rs->{comment_area} = 1 - $rs->{comment_area};
						}

					} else {
#								$rs->{body} .= sprintf("[0x%02X]",$val);
						if ($through_mode == 1) {
							$chrstr = "";
						} else {
							$chrstr = $char_code_table{$charset_type}{$val}{val};
						}
						if ($chrstr ne "") {
							$rs->{body} .= $chrstr;
						} else {
							$rs->{body} .= sprintf("%c",$val);
						}

					}
				}
			}
		}

		$rs->{subpos}++;

		if ($val == 0) {
			# end of line
			if ($through_mode == 1) {
				printf("%d %s\r",$rs->{line_num},$rs->{body});
			} else {
				printf("%d %s\n",$rs->{line_num},$rs->{body});
			}
			# clear val
			$rs->{mode} = 1;
			$rs->{quoted} = 0;
			$rs->{data_area} = 0;
			$rs->{comment_area} = 0;
			$rs->{subpos} = 0;
			$cmd = "";
		}
	}

}

sub set_basic_code_table()
{
	my($tbl)=@_;
	my $fh;
	my $key;
	my @arr;
	my $type = "";
	my $target = 0;

	if (! open($fh,$app_path."basic_code.dat")) {
		print STDERR "basic_code.dat open error.\n";
		return 1;
	}
	while (my $line = readline($fh)) {
		$line =~ s/^\s*(.+)\s*$/$1/;

		next if ($line =~ /^#/);

		chomp($line);

		# basic type
		if ($line =~ /^\[(.*)\]$/) {
			$type = $1;
			if ($type eq "" || $type eq $basic_type) {
				$target = 1;
			} else {
				$target = 0;
			}
		} else {
			next if ($target != 1);

			@arr = split(/\s*,\s*/,$line);

			next if ($arr[0] eq "");

			$key = hex($arr[0]);

			$tbl->{$key}{val} =$arr[1];
			$tbl->{$key}{attr}=$arr[2];
		}
#		printf STDERR "%d %s\n",$key,$tbl->{$key}{val};
	}
	close($fh);

	return 0;
}

sub set_char_code_table()
{
	my($tbl)=@_;
	my $fh;
	my $key;
	my @arr;
	my $charset = "";

	if (! open($fh,$app_path."char_code.dat")) {
		print STDERR "char_code.dat open error.\n";
		return 1;
	}
	while (my $line = readline($fh)) {
		$line =~ s/^\s*(.+)\s*$/$1/;

		next if ($line =~ /^#/);

		chomp($line);

		# charset type code
		if ($line =~ /^\[(.+)\]$/) {
			$charset = $1;
			if (!exists($tbl->{$charset})) {
				$tbl->{$charset} = {};
			}
#		printf STDERR "%s\n",$charset;
		} else {
			next if ($charset eq "");

			@arr = split(/\s*,\s*/,$line);

			next if ($arr[0] eq "");

			next if ($arr[2] ne "" && $arr[2] ne ">");

			$key = hex($arr[0]);

			next if ($key > 255);

			$arr[1] =~ s/^\'(.+)\'$/$1/;

			$tbl->{$charset}{$key}{val} =$arr[1];

#		printf STDERR "%d %s\n",$key,$tbl->{$charset}{$key}{val};
		}

	}
	close($fh);

	return 0;
}
