#!/usr/bin/perl
# encode code
# V0.4.3
#
# Copyright (c) 2012, Sasaji
#use utf8;
use strict;
use Getopt::Std;
use File::Basename;

our %opts = ();
our $charset_type = "0";	# 
our $through_mode = 1;		# 
our $cr_code = "\r";

our %char_code_table = ();
our %char_code_index = ();

our $app_path=dirname($0)."/";

getopts("i:n",\%opts);

if (exists($opts{"n"})) {
	$cr_code = "\n";
}
if ($opts{"i"} =~ /^[01]$/) {
	$through_mode = 0;
	$charset_type = $opts{"i"};
}

if (@ARGV <= 0) {
	&usage();
	exit 0;
}

if (set_char_code_table(\%char_code_table)) {
	exit 1;
}
foreach my $key (keys %char_code_table) {
	$char_code_index{$key}={} if (! defined($char_code_index{$key}));
	set_char_code_index($char_code_table{$key}, $char_code_index{$key});
}

foreach my $infile (@ARGV) {
	&main($infile);
}

# $k=getc(STDIN);

exit 0;

#####
sub usage() {
	my $prog_name=basename($0);

	print "$prog_name [options] file ...\n";
	print " Encode char code."."\n";
	print '  options:'."\n";
	print '  -i 0   : convert file(s) using interace mode charset(utf-8).'."\n";
	print '           zenkaku kata/hira -> hankaku kata/hira'."\n";
	print '  -i 1   : convert file(s) using non-interace mode charset(utf-8).'."\n";
}

#####

sub main() {
	my($infile) = @_;
	my $fh;

	my $buf="";
	my $buf1="";
	my $buflen = 0;

	if (! open($fh,$infile)) {
		print STDERR $infile.' cannot open.'."\n";
		return;
	}
	binmode($fh);

	my %parse_state;
	my $val = 0;

	&init_data(\%parse_state);

	binmode(STDOUT);
	my $len = 0;
	do {
		$buflen = read($fh,$buf1,2048);
		$buf .= $buf1;
		$len = length($buf);
		while($len >= 768 || $buflen <= 0) {
			$parse_state{file_line}++;
			$buf =~ s/^[\s\r\n]*//;
			last if ($buf eq "");

			if (&encode_data(\$buf, \%parse_state) != 0) {
				$buf =~ s/^.*?[\r\n]+//;
			}
			$len = length($buf);
		}
	} while ($buf ne "" || $buflen > 0);
	close($fh);

#	print "end\n";
}

#################
sub init_data() {
	my($rs)=@_;

	$rs->{mode} = 0;
	$rs->{file_line} = 0;		# 

	$rs->{subpos} = 0;			# 

	$rs->{body} = "";			# 

}

# parse per line
sub encode_data() {
	my($rline, $rs)=@_;
	my $statement="";
	my $new_state="";

	my $len_state = 0;

	$rs->{subpos} = 0;			# 

	$rs->{body} = "";			# 

	# parse statement
	$len_state = length($$rline);
	$new_state = "";
	while($rs->{subpos} < $len_state) {
		my $ch = substr($$rline, $rs->{subpos}, 1);
		my $find = 0;

		if ($ch =~ /^[\r\n]/) {
			last;
		}

		if ($through_mode == 0) {
			# convert char code
			if (exists($char_code_index{$charset_type}{$ch})) {
				foreach my $cmd (@{$char_code_index{$charset_type}{$ch}}) {
					my $cmdlen = length($cmd->{key});
					my $cmdpos = index($$rline, $cmd->{key}, $rs->{subpos});
					if ($cmdpos == $rs->{subpos}) {
						# match
						$new_state .= $cmd->{data}{val};
						$find = 1;
						$rs->{subpos} += $cmdlen;
						last;
					} 
				}
			}
		}
		if ($find != 1) {
			$new_state .= $ch;
			$rs->{subpos}++;
		}
	}

	$rs->{body} = $new_state.$cr_code;

	print $rs->{body};

	$$rline = substr($$rline, $rs->{subpos});

	return 0;
}


sub set_char_code_index()
{
	my($tbl, $idx)=@_;

	foreach my $key (sort keys %{$tbl}) {
		my $ch = substr($key, 0, 1);
		$idx->{$ch} = [] if (!exists($idx->{$ch}));
		unshift(@{$idx->{$ch}}, { key => $key, data => $tbl->{$key} });

#		printf STDERR "%s\n",$key;
	}
	return $idx;
}

sub set_char_code_table()
{
	my($tbl)=@_;
	my $fh;
	my $key;
	my $val;
	my @arr;
	my $charset = "";

	if (! open($fh,$app_path."char_code.dat")) {
		print STDERR "char_code.dat open error.\n";
		return 1;
	}
	while (my $line = readline($fh)) {
		$line =~ s/^\s*(.+)\s*$/$1/;

		next if ($line =~ /^#/);

		chomp($line);

		# charset type code
		if ($line =~ /^\[(.+)\]$/) {
			$charset = $1;
			if (!exists($tbl->{$charset})) {
				$tbl->{$charset} = {};
			}
#		printf STDERR "%s\n",$charset;
		} else {
			next if ($charset eq "");

			@arr = split(/\s*,\s*/,$line);

			next if ($arr[0] eq "");

			next if ($arr[2] ne "" && $arr[2] ne "<");

			$arr[1] =~ s/^\'(.+)\'$/$1/;
			$key = $arr[1];
			$val = hex($arr[0]);
			if ($val >= 256) {
				$tbl->{$charset}{$key}{val} = pack("C2",($val >> 8) & 0xff,$val & 0xff);
			} else {
				$tbl->{$charset}{$key}{val} = pack("C",$val & 0xff);
			}
#		printf STDERR "%s %s\n",$key,$tbl->{$charset}{$key}{val};
		}

	}
	close($fh);

	return 0;
}
